@@ -19,9 +19,9 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  *  USA
  *
- *  $Id: patchlevel.h,v 1.1.1.1 2002/09/12 01:08:41 cry0gen Exp $
+ *  $Id: patchlevel.h 843 2007-02-20 20:16:19Z stu $
  */
 
 #ifndef PATCHLEVEL
-#define PATCHLEVEL  "hybrid-7+oftc1.1.0"
+#define PATCHLEVEL   "hybrid-7.2.2+oftc1.5.5"
 #endif
