@@ -19,18 +19,18 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  *  USA
  *
- *  $Id: rsa.h,v 1.1.1.1 2002/09/12 01:08:38 cry0gen Exp $
+ *  $Id: rsa.h 588 2006-05-29 18:45:22Z stu $
  */
 #ifndef INCLUDED_rsa_h
-#define INCLUDED_rsa_h 1
+#define INCLUDED_rsa_h
+
+#include "setup.h"
 
 #ifdef HAVE_LIBCRYPTO
 extern void report_crypto_errors(void);
 extern int verify_private_key(void);
-extern int generate_challenge( char ** r_challenge, char ** r_response, RSA * key );
-extern int crypt_data(char **out, char *in, int len, char *key);
-extern int decrypt_data(char **out, char *in, int len, char *key);
-extern int get_randomness( unsigned char * buf, int length );
+extern int generate_challenge(char **, char **, RSA *);
+extern int get_randomness(unsigned char *, int);
 #endif
 
 #endif /* INCLUDED_rsa_h */
