@@ -19,7 +19,7 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  *  USA
  *
- *  $Id: m_message.c,v 1.1 2002/09/12 02:16:17 cry0gen Exp $
+ *  $Id: m_message.c 722 2006-11-22 18:29:54Z stu $
  */
 
 #include "stdinc.h"
@@ -34,17 +34,12 @@
 #include "msg.h"
 #include "parse.h"
 #include "modules.h"
-
 #include "channel.h"
 #include "channel_mode.h"
-#include "vchannel.h"
 #include "irc_string.h"
 #include "hash.h"
-#include "class.h"
-#include "msg.h"
 #include "packet.h"
 
-
 struct entity
 {
   void *ptr;
@@ -52,7 +47,7 @@ struct entity
   int flags;
 };
 
-static int build_target_list(int p_or_n, char *command,
+static int build_target_list(int p_or_n, const char *command,
                              struct Client *client_p,
                              struct Client *source_p,
                              char *nicks_channels, char *text);
@@ -73,47 +68,51 @@ static int ntargets = 0;
 
 static int duplicate_ptr(void *);
 
-static void m_message(int, char *, struct Client *,
+static void m_message(int, const char *, struct Client *,
                       struct Client *, int, char **);
 
 static void m_privmsg(struct Client *, struct Client *, int, char **);
 static void m_notice(struct Client *, struct Client *, int, char **);
 
-static void msg_channel(int p_or_n, char *command,
+static void msg_channel(int p_or_n, const char *command,
                         struct Client *client_p,
                         struct Client *source_p,
                         struct Channel *chptr, char *text);
 
-static void msg_channel_flags(int p_or_n, char *command,
+static void msg_channel_flags(int p_or_n, const char *command,
                               struct Client *client_p,
                               struct Client *source_p,
                               struct Channel *chptr, int flags, char *text);
 
-static void msg_client(int p_or_n, char *command,
+static void msg_client(int p_or_n, const char *command,
                        struct Client *source_p, struct Client *target_p,
                        char *text);
 
-static void handle_opers(int p_or_n, char *command,
-                         struct Client *client_p,
-                         struct Client *source_p, char *nick, char *text);
+static void handle_special(int p_or_n, const char *command,
+			   struct Client *client_p,
+			   struct Client *source_p, char *nick, char *text);
 
 struct Message privmsg_msgtab = {
   "PRIVMSG", 0, 0, 1, 0, MFLG_SLOW | MFLG_UNREG, 0L,
-  {m_unregistered, m_privmsg, m_privmsg, m_privmsg}
+  {m_unregistered, m_privmsg, m_privmsg, m_ignore, m_privmsg, m_ignore}
 };
 
 struct Message notice_msgtab = {
   "NOTICE", 0, 0, 1, 0, MFLG_SLOW, 0L,
-  {m_unregistered, m_notice, m_notice, m_notice}
+  {m_unregistered, m_notice, m_notice, m_ignore, m_notice, m_ignore}
 };
 
 #ifndef STATIC_MODULES
+struct Callback *client_message;
+struct Callback *channel_message;
 
 void
 _modinit(void)
 {
   mod_add_cmd(&privmsg_msgtab);
   mod_add_cmd(&notice_msgtab);
+  client_message = register_callback("client_message", NULL);
+  channel_message = register_callback("channel_message", NULL);
 }
 
 void
@@ -123,7 +122,7 @@ _moddeinit(void)
   mod_del_cmd(&notice_msgtab);
 }
 
-const char *_version = "$Revision: 1.1 $";
+const char *_version = "$Revision: 722 $";
 #endif
 
 /*
@@ -149,21 +148,21 @@ const char *_version = "$Revision: 1.1 $";
 #define NOTICE  1
 
 static void
-m_privmsg(struct Client *client_p,
-          struct Client *source_p, int parc, char *parv[])
+m_privmsg(struct Client *client_p, struct Client *source_p,
+          int parc, char *parv[])
 {
   /* servers have no reason to send privmsgs, yet sometimes there is cause
    * for a notice.. (for example remote kline replies) --fl_
    */
-  if (!IsPerson(source_p))
+  if (!IsClient(source_p))
     return;
 
   m_message(PRIVMSG, "PRIVMSG", client_p, source_p, parc, parv);
 }
 
 static void
-m_notice(struct Client *client_p,
-         struct Client *source_p, int parc, char *parv[])
+m_notice(struct Client *client_p, struct Client *source_p,
+         int parc, char *parv[])
 {
   m_message(NOTICE, "NOTICE", client_p, source_p, parc, parv);
 }
@@ -176,31 +175,41 @@ m_notice(struct Client *client_p,
  *		- pointer to channel
  */
 static void
-m_message(int p_or_n,
-          char *command,
-          struct Client *client_p,
+m_message(int p_or_n, const char *command, struct Client *client_p,
           struct Client *source_p, int parc, char *parv[])
 {
   int i;
 
-  if (parc < 2 || *parv[1] == '\0')
+  if (parc < 2 || EmptyString(parv[1]))
   {
     if (p_or_n != NOTICE)
-      sendto_one(source_p, form_str(ERR_NORECIPIENT), me.name,
-                 source_p->name, command);
+      sendto_one(source_p, form_str(ERR_NORECIPIENT),
+                 ID_or_name(&me, client_p),
+                 ID_or_name(source_p, client_p), command);
     return;
   }
 
-  if (parc < 3 || *parv[2] == '\0')
+  if (parc < 3 || EmptyString(parv[2]))
   {
     if (p_or_n != NOTICE)
-      sendto_one(source_p, form_str(ERR_NOTEXTTOSEND), me.name,
-                 source_p->name);
+      sendto_one(source_p, form_str(ERR_NOTEXTTOSEND),
+                 ID_or_name(&me, client_p),
+                 ID_or_name(source_p, client_p));
     return;
   }
 
   /* Finish the flood grace period... */
-  if(MyClient(source_p) && !IsFloodDone(source_p))
+  if (MyClient(source_p) && !IsFloodDone(source_p))
+#if 0
+ &&
+        irccmp(source_p->name, parv[1]) != 0) /* some dumb clients msg/notice themself
+                                                 to determine lag to the server BEFORE
+                                                 sending JOIN commands, and then flood
+                                                 off because they left gracemode. -wiz */
+	/*
+	 * Not our problem if they do this.    -Michael
+         */
+#endif
     flood_endgrace(source_p);
 
   if (build_target_list(p_or_n, command, client_p, source_p, parv[1],
@@ -236,8 +245,7 @@ m_message(int p_or_n,
   }
 }
 
-/*
- * build_target_list
+/* build_target_list()
  *
  * inputs	- pointer to given client_p (server)
  *		- pointer to given source (oper/client etc.)
@@ -253,21 +261,19 @@ m_message(int p_or_n,
  *		  to disambiguate.
  *
  */
-
 static int
-build_target_list(int p_or_n, char *command, struct Client *client_p,
-                  struct Client *source_p, char *nicks_channels,
-                  char *text)
+build_target_list(int p_or_n, const char *command, struct Client *client_p,
+                  struct Client *source_p, char *nicks_channels, char *text)
 {
   int type;
-  char *p, *nick, *target_list, ncbuf[BUFSIZE];
-  struct Channel *chptr=NULL;
-  struct Client *target_p;
+  char *p, *nick, *target_list, ncbuf[IRCD_BUFSIZE];
+  struct Channel *chptr = NULL;
+  struct Client *target_p = NULL;
 
   /* Sigh, we can't mutilate parv[1] incase we need it to send to a hub */
   if (!ServerInfo.hub && (uplink != NULL) && IsCapable(uplink, CAP_LL))
   {
-    strncpy(ncbuf, nicks_channels, BUFSIZE);
+    strlcpy(ncbuf, nicks_channels, sizeof(ncbuf));
     target_list = ncbuf;
   }
   else
@@ -276,7 +282,7 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
   ntargets = 0;
 
   for (nick = strtoken(&p, target_list, ","); nick;
-       nick = strtoken(&p, (char *)NULL, ","))
+       nick = strtoken(&p, NULL, ","))
   {
     char *with_prefix;
     /*
@@ -287,7 +293,7 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
     if (IsChanPrefix(*nick))
     {
       /* ignore send of local channel to a server (should not happen) */
-      if (IsServer(client_p) && *nick == '&')
+      if (*nick == '&' && IsServer(client_p))
         continue;
 
       if ((chptr = hash_find_channel(nick)) != NULL)
@@ -295,11 +301,13 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
         if (!duplicate_ptr(chptr))
         {
           if (ntargets >= ConfigFileEntry.max_targets)
-	    {
-	      sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
-			 me.name, source_p->name, nick);
-	      return (1);
-	    }
+          {
+            sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
+                       ID_or_name(&me, client_p),
+                       ID_or_name(source_p, client_p), nick,
+		       ConfigFileEntry.max_targets);
+            return (1);
+          }
           targets[ntargets].ptr = (void *)chptr;
           targets[ntargets++].type = ENTITY_CHANNEL;
         }
@@ -309,21 +317,24 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
         if (!ServerInfo.hub && (uplink != NULL) && IsCapable(uplink, CAP_LL))
           return -1;
         else if (p_or_n != NOTICE)
-          sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name,
-                     source_p->name, nick);
+          sendto_one(source_p, form_str(ERR_NOSUCHNICK),
+                     ID_or_name(&me, client_p),
+                     ID_or_name(source_p, client_p), nick);
       }
       continue;
     }
 
     /* look for a privmsg to another client */
-    if ((target_p = find_person(nick)) != NULL)
+    if ((target_p = find_person(client_p, nick)) != NULL)
     {
       if (!duplicate_ptr(target_p))
       {
         if (ntargets >= ConfigFileEntry.max_targets)
 	  {
 	    sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
-		       me.name, source_p->name, nick);
+                       ID_or_name(&me, client_p),
+                       ID_or_name(source_p, client_p), nick,
+		       ConfigFileEntry.max_targets);
 	    return (1);
 	  }
         targets[ntargets].ptr = (void *)target_p;
@@ -338,14 +349,16 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
     type = 0;
     with_prefix = nick;
     /*  allow %+@ if someone wants to do that */
-    for (;;)
+    for (; ;)
     {
       if (*nick == '@')
-        type |= MODE_CHANOP;
+        type |= CHFL_CHANOP;
+#ifdef HALFOPS
       else if (*nick == '%')
-        type |= MODE_CHANOP | MODE_HALFOP;
+        type |= CHFL_CHANOP | CHFL_HALFOP;
+#endif
       else if (*nick == '+')
-        type |= MODE_CHANOP | MODE_HALFOP | MODE_VOICE;
+        type |= CHFL_CHANOP | CHFL_HALFOP | CHFL_VOICE;
       else
         break;
       nick++;
@@ -355,68 +368,78 @@ build_target_list(int p_or_n, char *command, struct Client *client_p,
     {
       /* suggested by Mortiis */
       if (*nick == '\0')      /* if its a '\0' dump it, there is no recipient */
-	{
-	  sendto_one(source_p, form_str(ERR_NORECIPIENT),
-		     me.name, source_p->name, command);
-	  continue;
-	}
+      {
+        sendto_one(source_p, form_str(ERR_NORECIPIENT),
+                   ID_or_name(&me, client_p),
+                   ID_or_name(source_p, client_p), command);
+        continue;
+      }
 
       /* At this point, nick+1 should be a channel name i.e. #foo or &foo
        * if the channel is found, fine, if not report an error
        */
 
       if ((chptr = hash_find_channel(nick)) != NULL)
-	{
-	  if(!is_any_op(chptr, source_p) && !is_voiced(chptr, source_p))
-	    {
-	      sendto_one(source_p, form_str(ERR_CHANOPRIVSNEEDED), me.name,
-			 source_p->name, with_prefix);
-	      return(-1);
-	    }
+      {
+        if (!has_member_flags(find_channel_link(source_p, chptr),
+                              CHFL_CHANOP|CHFL_HALFOP|CHFL_VOICE))
+        {
+          sendto_one(source_p, form_str(ERR_CHANOPRIVSNEEDED),
+                     ID_or_name(&me, client_p),
+                     ID_or_name(source_p, client_p), with_prefix);
+          return(-1);
+        }
 
-	  if (!duplicate_ptr(chptr))
-	    {
-	      if (ntargets >= ConfigFileEntry.max_targets)
-		{
-		  sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
-			     me.name, source_p->name, nick);
-		  return (1);
-		}
-	      targets[ntargets].ptr = (void *)chptr;
-	      targets[ntargets].type = ENTITY_CHANOPS_ON_CHANNEL;
-	      targets[ntargets++].flags = type;
-	    }
-	}
+        if (!duplicate_ptr(chptr))
+        {
+          if (ntargets >= ConfigFileEntry.max_targets)
+          {
+	    sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
+                       ID_or_name(&me, client_p),
+                       ID_or_name(source_p, client_p), nick,
+		       ConfigFileEntry.max_targets);
+            return(1);
+          }
+          targets[ntargets].ptr = (void *)chptr;
+          targets[ntargets].type = ENTITY_CHANOPS_ON_CHANNEL;
+          targets[ntargets++].flags = type;
+        }
+      }
       else
-	{
-	  if (!ServerInfo.hub && (uplink != NULL) && IsCapable(uplink, CAP_LL))
-	    return -1;
-	  else if (p_or_n != NOTICE)
-	    sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name,
-		       source_p->name, nick);
-	}
+      {
+        if (!ServerInfo.hub && (uplink != NULL) && IsCapable(uplink, CAP_LL))
+          return -1;
+        else if (p_or_n != NOTICE)
+          sendto_one(source_p, form_str(ERR_NOSUCHNICK),
+                     ID_or_name(&me, client_p),
+                     ID_or_name(source_p, client_p), nick);
+      }
       continue;
     }
 
-    if(IsOper(source_p) && ((*nick == '$') || strchr(nick, '@')))
+    if ((*nick == '$') || strchr(nick, '@') != NULL)
     {
-      handle_opers(p_or_n, command, client_p, source_p, nick, text);
+      handle_special(p_or_n, command, client_p, source_p, nick, text);
     }
     else
     {
       if (!ServerInfo.hub && (uplink != NULL) && IsCapable(uplink, CAP_LL))
         return -1;
-      else if(p_or_n != NOTICE)
-        sendto_one(source_p, form_str(ERR_NOSUCHNICK),
-	           me.name, source_p->name, nick);
+      else if (p_or_n != NOTICE)
+      {
+        if (!IsDigit(*nick) || MyClient(source_p))
+	  sendto_one(source_p, form_str(ERR_NOSUCHNICK),
+		     ID_or_name(&me, client_p),
+		     ID_or_name(source_p, client_p), nick);
+      }
     }
     /* continue; */
   }
-  return (1);
+
+  return(1);
 }
 
-/*
- * duplicate_ptr
+/* duplicate_ptr()
  *
  * inputs	- pointer to check
  *		- pointer to table of entities
@@ -429,14 +452,17 @@ static int
 duplicate_ptr(void *ptr)
 {
   int i;
+
   for (i = 0; i < ntargets; i++)
+  {
     if (targets[i].ptr == ptr)
-      return YES;
-  return NO;
+      return(1);
+  }
+
+  return(0);
 }
 
-/*
- * msg_channel
+/* msg_channel()
  *
  * inputs	- flag privmsg or notice
  * 		- pointer to command "PRIVMSG" or "NOTICE"
@@ -447,52 +473,50 @@ duplicate_ptr(void *ptr)
  * side effects	- message given channel
  */
 static void
-msg_channel(int p_or_n, char *command,
-            struct Client *client_p,
+msg_channel(int p_or_n, const char *command, struct Client *client_p,
             struct Client *source_p, struct Channel *chptr, char *text)
 {
-  struct Channel *vchan = NULL;
-  char *chname = NULL;
   int result;
 
-  chname = RootChan(chptr)->chname;
-
-#ifdef VCHANS
-  if (HasVchans(chptr))
-    vchan = map_vchan(chptr, source_p);
-#endif
-  
-  if (!vchan)
-    vchan = chptr;
-
   if (MyClient(source_p))
   {
     /* idle time shouldnt be reset by notices --fl */
-    if ((p_or_n != NOTICE) && source_p->user)
-      source_p->user->last = CurrentTime;
+    if (p_or_n != NOTICE)
+      source_p->localClient->last = CurrentTime;
   }
 
+#ifndef STATIC_MODULES
+  execute_callback(channel_message, source_p, chptr, text);
+#endif
+
   /* chanops and voiced can flood their own channel with impunity */
-  if ((result = can_send(vchan, source_p)))
+  result = can_send(chptr, source_p, NULL);
+  if (result > 0)
   {
     if (result == CAN_SEND_OPV ||
-        !flood_attack_channel(p_or_n, source_p, vchan, chname))
+        !flood_attack_channel(p_or_n, source_p, chptr, chptr->chname))
     {
-      if(vchan->mode.mode & MODE_NOCOLOR && msg_has_colors(text))
-          text = strip_color(text);
-      sendto_channel_butone(client_p, source_p, vchan, command, ":%s", text);
+      if(chptr->mode.mode & MODE_NOCOLOR && msg_has_colors(text))
+        text = strip_color(text);
+      sendto_channel_butone(client_p, source_p, chptr, command, ":%s", text);
     }
   }
   else
   {
     if (p_or_n != NOTICE)
+    {
       sendto_one(source_p, form_str(ERR_CANNOTSENDTOCHAN),
-                 me.name, source_p->name, chname);
+                 ID_or_name(&me, client_p),
+                 ID_or_name(source_p, client_p), chptr->chname);
+      if(result == CAN_SEND_ONLY_IF_REG)
+        sendto_one(source_p, form_str(ERR_CANTSENDREGONLY),
+            ID_or_name(&me, client_p),
+            ID_or_name(source_p, client_p), chptr->chname);
+    }
   }
 }
 
-/*
- * msg_channel_flags
+/* msg_channel_flags()
  *
  * inputs	- flag 0 if PRIVMSG 1 if NOTICE. RFC 
  *		  say NOTICE must not auto reply
@@ -506,67 +530,63 @@ msg_channel(int p_or_n, char *command,
  * side effects	- message given channel either chanop or voice
  */
 static void
-msg_channel_flags(int p_or_n, char *command, struct Client *client_p,
+msg_channel_flags(int p_or_n, const char *command, struct Client *client_p,
                   struct Client *source_p, struct Channel *chptr,
                   int flags, char *text)
 {
-  struct Channel *vchan = NULL;
-  char *chname = NULL;
   int type;
   char c;
 
-  if (flags & MODE_VOICE)
+  if (flags & CHFL_VOICE)
   {
-    type = ONLY_CHANOPS_HALFOPS_VOICED;
+    type = CHFL_VOICE|CHFL_HALFOP|CHFL_CHANOP;
     c = '+';
   }
-  else if (flags & MODE_HALFOP)
+#ifdef HALFOPS
+  else if (flags & CHFL_HALFOP)
   {
-    type = ONLY_CHANOPS_HALFOPS;
+    type = CHFL_HALFOP|CHFL_CHANOP;
     c = '%';
-  }
+   }
+#endif
   else
   {
-    type = ONLY_CHANOPS;
+    type = CHFL_CHANOP;
     c = '@';
   }
 
-  chname = RootChan(chptr)->chname;
-
-#ifdef VCHANS
-  if (HasVchans(chptr))
-    vchan = map_vchan(chptr, source_p);
-#endif
-
-  if (!vchan)
-    vchan = chptr;
-
   if (MyClient(source_p))
   {
     /* idletime shouldnt be reset by notice --fl */
-    if ((p_or_n != NOTICE) && source_p->user)
-      source_p->user->last = CurrentTime;
+    if (p_or_n != NOTICE)
+      source_p->localClient->last = CurrentTime;
+
+    sendto_channel_local_butone(source_p, type, chptr, ":%s!%s@%s %s %c%s :%s",
+                                source_p->name, source_p->username,
+                                source_p->host, command, c, chptr->chname, text);
+  }
+  else
+  {
+    /*
+     * another good catch, lee.  we never would echo to remote clients anyway,
+     * so use slightly less intensive sendto_channel_local()
+     */
+    sendto_channel_local(type, YES, chptr, ":%s!%s@%s %s %c%s :%s",
+                         source_p->name, source_p->username,
+                         source_p->host, command, c, chptr->chname, text);
   }
 
-  sendto_channel_local(type, vchan, ":%s!%s@%s %s %c%s :%s",
-                       source_p->name, source_p->username,
-                       source_p->host, command, c, chname, text);
-
-  if (chptr->chname[0] == '&')
+  if (chptr->chname[0] != '#')
     return;
 
-  sendto_channel_remote(source_p, client_p, type, CAP_CHW, CAP_UID, vchan,
-                ":%s %s %c%s :%s", source_p->name, command, c,
-                vchan->chname, text);
-  sendto_channel_remote(source_p, client_p, type, CAP_CHW|CAP_UID, NOCAPS, vchan,
-                ":%s %s %c%s :%s", ID(source_p), command, c,
-                vchan->chname, text);
-
+  sendto_channel_remote(source_p, client_p, type, CAP_CHW, CAP_TS6, chptr,
+                ":%s %s %c%s :%s", source_p->name, command, c, chptr->chname, text);
+  sendto_channel_remote(source_p, client_p, type, CAP_CHW|CAP_TS6, NOCAPS, chptr,
+                ":%s %s %c%s :%s", ID(source_p), command, c, chptr->chname, text);
   /* non CAP_CHW servers? */
 }
 
-/*
- * msg_client
+/* msg_client()
  *
  * inputs	- flag 0 if PRIVMSG 1 if NOTICE. RFC 
  *		  say NOTICE must not auto reply
@@ -578,61 +598,68 @@ msg_channel_flags(int p_or_n, char *command, struct Client *client_p,
  * side effects	- message given channel either chanop or voice
  */
 static void
-msg_client(int p_or_n, char *command,
-           struct Client *source_p, struct Client *target_p, char *text)
+msg_client(int p_or_n, const char *command, struct Client *source_p,
+           struct Client *target_p, char *text)
 {
   if (MyClient(source_p))
   {
-    /* reset idle time for message only if its not to self 
-     * and its not a notice */
-    if ((p_or_n != NOTICE) && (source_p != target_p) && source_p->user)
-      source_p->user->last = CurrentTime;
+    /*
+     * reset idle time for message only if its not to self 
+     * and its not a notice
+     * NOTE: Normally we really should reset their idletime
+     * if the are messaging themselves, but we have to go
+     * this way in order to prevent them to trick out
+     * idle-klines.
+     */
+    if ((p_or_n != NOTICE) && (source_p != target_p))
+      source_p->localClient->last = CurrentTime;
   }
 
-  if (MyConnect(source_p) && (p_or_n != NOTICE) &&
-      target_p->user && target_p->user->away)
+#ifndef STATIC_MODULES
+  execute_callback(client_message, source_p, target_p, text);
+#endif
+
+  if (MyConnect(source_p) && (p_or_n != NOTICE) && target_p->away)
     sendto_one(source_p, form_str(RPL_AWAY), me.name,
-               source_p->name, target_p->name, target_p->user->away);
+               source_p->name, target_p->name, target_p->away);
 
   if (MyClient(target_p))
   {
-    /* XXX Controversial? allow opers always to send through a +g */
     if (!IsServer(source_p) && IsSetCallerId(target_p))
     {
       /* Here is the anti-flood bot/spambot code -db */
-      if (accept_message(source_p, target_p) || IsOper(source_p))
+      if (source_p == target_p || accept_message(source_p, target_p) ||
+         (IsOper(source_p) && (ConfigFileEntry.opers_bypass_callerid == 1)))
       {
         sendto_one(target_p, ":%s!%s@%s %s %s :%s",
-                   source_p->name,
-                   source_p->username,
+                   source_p->name, source_p->username,
                    source_p->host, command, target_p->name, text);
       }
       else
       {
         /* check for accept, flag recipient incoming message */
         if (p_or_n != NOTICE)
-          sendto_anywhere(source_p, target_p,
-                          "NOTICE %s :*** I'm in +g mode (server side ignore).",
-                          source_p->name);
+	  sendto_one(source_p, form_str(RPL_TARGUMODEG),
+		     ID_or_name(&me, source_p->from),
+		     ID_or_name(source_p, source_p->from), target_p->name);
 
         if ((target_p->localClient->last_caller_id_time +
              ConfigFileEntry.caller_id_wait) < CurrentTime)
         {
           if (p_or_n != NOTICE)
-            sendto_anywhere(source_p, target_p,
-                            "NOTICE %s :*** I've been informed you messaged me.",
-                            source_p->name);
+	    sendto_one(source_p, form_str(RPL_TARGNOTIFY),
+		       ID_or_name(&me, source_p->from),
+		       ID_or_name(source_p, source_p->from), target_p->name);
 
-          sendto_one(target_p,
-                     ":%s NOTICE %s :*** Client %s [%s@%s] is messaging you and you are +g",
+          sendto_one(target_p, form_str(RPL_UMODEGMSG),
                      me.name, target_p->name,
-                     source_p->name, source_p->username, source_p->host);
+                     get_client_name(source_p, HIDE_IP));
 
           target_p->localClient->last_caller_id_time = CurrentTime;
 
         }
         /* Only so opers can watch for floods */
-        (void)flood_attack_client(p_or_n, source_p, target_p);
+        flood_attack_client(p_or_n, source_p, target_p);
       }
     }
     else
@@ -656,11 +683,10 @@ msg_client(int p_or_n, char *command,
            && !flood_attack_client(p_or_n, source_p, target_p)))
     sendto_anywhere(target_p, source_p, "%s %s :%s", command, target_p->name,
                     text);
-  return;
 }
 
-/*
- * flood_attack_client
+/* flood_attack_client()
+ *
  * inputs       - flag 0 if PRIVMSG 1 if NOTICE. RFC
  *                say NOTICE must not auto reply
  *              - pointer to source Client 
@@ -675,7 +701,7 @@ flood_attack_client(int p_or_n, struct Client *source_p,
   int delta;
 
   if (GlobalSetOptions.floodcount && MyConnect(target_p)
-      && IsClient(source_p))
+      && IsClient(source_p) && !IsConfCanFlood(source_p))
   {
     if ((target_p->localClient->first_received_message_time + 1)
         < CurrentTime)
@@ -696,30 +722,30 @@ flood_attack_client(int p_or_n, struct Client *source_p,
     {
       if (target_p->localClient->flood_noticed == 0)
       {
-        sendto_gnotice_flags(FLAGS_BOTS, L_OPER, me.name, &me, NULL,
-                             "Possible Flooder %s [%s@%s] on %s target: %s",
-                             source_p->name, source_p->username,
-                             source_p->host,
-                             source_p->user->server, target_p->name);
+        sendto_realops_flags(UMODE_BOTS, L_ALL,
+                             "Possible Flooder %s on %s target: %s",
+                             get_client_name(source_p, HIDE_IP),
+                             source_p->servptr->name, target_p->name);
         target_p->localClient->flood_noticed = 1;
         /* add a bit of penalty */
         target_p->localClient->received_number_of_privmsgs += 2;
       }
+
       if (MyClient(source_p) && (p_or_n != NOTICE))
         sendto_one(source_p,
                    ":%s NOTICE %s :*** Message to %s throttled due to flooding",
                    me.name, source_p->name, target_p->name);
-      return 1;
+      return(1);
     }
     else
       target_p->localClient->received_number_of_privmsgs++;
   }
 
-  return 0;
+  return(0);
 }
 
-/*
- * flood_attack_channel
+/* flood_attack_channel()
+ *
  * inputs       - flag 0 if PRIVMSG 1 if NOTICE. RFC
  *                says NOTICE must not auto reply
  *              - pointer to source Client 
@@ -733,7 +759,7 @@ flood_attack_channel(int p_or_n, struct Client *source_p,
 {
   int delta;
 
-  if (GlobalSetOptions.floodcount)
+  if (GlobalSetOptions.floodcount && !IsConfCanFlood(source_p))
   {
     if ((chptr->first_received_message_time + 1) < CurrentTime)
     {
@@ -743,21 +769,20 @@ flood_attack_channel(int p_or_n, struct Client *source_p,
       if (chptr->received_number_of_privmsgs <= 0)
       {
         chptr->received_number_of_privmsgs = 0;
-        chptr->flood_noticed = 0;
+        ClearFloodNoticed(chptr);
       }
     }
 
     if ((chptr->received_number_of_privmsgs >= GlobalSetOptions.floodcount)
-        || chptr->flood_noticed)
+        || IsSetFloodNoticed(chptr))
     {
-      if (chptr->flood_noticed == 0)
+      if (!IsSetFloodNoticed(chptr))
       {
-        sendto_gnotice_flags(FLAGS_BOTS, L_OPER, me.name, &me, NULL,
-                             "Possible Flooder %s [%s@%s] on %s target: %s",
-                             source_p->name, source_p->username,
-                             source_p->host,
-                             source_p->user->server, chptr->chname);
-        chptr->flood_noticed = 1;
+        sendto_realops_flags(UMODE_BOTS, L_ALL,
+                             "Possible Flooder %s on %s target: %s",
+                             get_client_name(source_p, HIDE_IP),
+                             source_p->servptr->name, chptr->chname);
+        SetFloodNoticed(chptr);
 
         /* Add a bit of penalty */
         chptr->received_number_of_privmsgs += 2;
@@ -766,35 +791,39 @@ flood_attack_channel(int p_or_n, struct Client *source_p,
         sendto_one(source_p,
                    ":%s NOTICE %s :*** Message to %s throttled due to flooding",
                    me.name, source_p->name, chname);
-      return 1;
+      return(1);
     }
     else
       chptr->received_number_of_privmsgs++;
   }
 
-  return 0;
+  return(0);
 }
 
-
-/*
- * handle_opers
+/* handle_special()
  *
  * inputs	- server pointer
  *		- client pointer
  *		- nick stuff to grok for opers
  *		- text to send if grok
  * output	- none
- * side effects	- all the traditional oper type messages are parsed here.
+ * side effects	- old style username@server is handled here for non opers
+ *		  opers are allowed username%hostname@server
+ *		  all the traditional oper type messages are also parsed here.
  *		  i.e. "/msg #some.host."
  *		  However, syntax has been changed.
  *		  previous syntax "/msg #some.host.mask"
  *		  now becomes     "/msg $#some.host.mask"
  *		  previous syntax of: "/msg $some.server.mask" remains
  *		  This disambiguates the syntax.
+ *
+ * XXX		  N.B. dalnet changed it to nick@server as have other servers.
+ *		  we will stick with tradition for now.
+ *		- Dianora
  */
 static void
-handle_opers(int p_or_n, char *command, struct Client *client_p,
-             struct Client *source_p, char *nick, char *text)
+handle_special(int p_or_n, const char *command, struct Client *client_p,
+	       struct Client *source_p, char *nick, char *text)
 {
   struct Client *target_p;
   char *host;
@@ -802,15 +831,111 @@ handle_opers(int p_or_n, char *command, struct Client *client_p,
   char *s;
   int count;
 
+  /*
+   * user[%host]@server addressed?
+   */
+  if ((server = strchr(nick, '@')) != NULL)
+  {
+    count = 0;
+
+    if ((host = strchr(nick, '%')) && !IsOper(source_p))
+    {
+      sendto_one(source_p, form_str(ERR_NOPRIVILEGES),
+                 ID_or_name(&me, client_p),
+                 ID_or_name(source_p, client_p));
+      return;
+    }
+
+    if ((target_p = find_server(server + 1)) != NULL)
+    {
+      if (!IsMe(target_p))
+      {
+	/*
+	 * Not destined for a user on me :-(
+	 */
+	sendto_one(target_p, ":%s %s %s :%s",
+                   ID_or_name(source_p, target_p->from),
+		   command, nick, text);
+	if ((p_or_n != NOTICE) && MyClient(source_p))
+	  source_p->localClient->last = CurrentTime;
+	return;
+      }
+
+      *server = '\0';
+
+      if (host != NULL)
+	*host++ = '\0';
+
+      /* Check if someones msg'ing opers@our.server */
+      if (strcmp(nick, "opers") == 0)
+      {
+	if (!IsOper(source_p))
+	  sendto_one(source_p, form_str(ERR_NOPRIVILEGES),
+                     ID_or_name(&me, client_p),
+                     ID_or_name(source_p, client_p));
+	else
+	  sendto_realops_flags(UMODE_ALL, L_ALL, "To opers: From: %s: %s",
+			       source_p->name, text);
+	return;
+      }
+
+      /*
+       * Look for users which match the destination host
+       * (no host == wildcard) and if one and one only is
+       * found connected to me, deliver message!
+       */
+      target_p = find_userhost(nick, host, &count);
+
+      if (target_p != NULL)
+      {
+	if (server != NULL)
+	  *server = '@';
+	if (host != NULL)
+	  *--host = '%';
+
+	if (count == 1)
+	{
+	  sendto_one(target_p, ":%s!%s@%s %s %s :%s",
+		     source_p->name, source_p->username, source_p->host,
+                     command, nick, text);
+	  if ((p_or_n != NOTICE) && MyClient(source_p))
+	    source_p->localClient->last = CurrentTime;
+	}
+	else
+	  sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
+                     ID_or_name(&me, client_p),
+                     ID_or_name(source_p, client_p), nick,
+		     ConfigFileEntry.max_targets);
+      }
+    }
+    else if (server && *(server+1) && (target_p == NULL))
+	sendto_one(source_p, form_str(ERR_NOSUCHSERVER),
+                   ID_or_name(&me, client_p),
+                   ID_or_name(source_p, client_p), server+1);
+    else if (server && (target_p == NULL))
+	sendto_one(source_p, form_str(ERR_NOSUCHNICK),
+                   ID_or_name(&me, client_p),
+                   ID_or_name(source_p, client_p), nick);
+    return;
+  }
+
+  if (!IsOper(source_p))
+  {
+    sendto_one(source_p, form_str(ERR_NOPRIVILEGES),
+               ID_or_name(&me, client_p),
+               ID_or_name(source_p, client_p));
+    return;
+  }
+
   /*
    * the following two cases allow masks in NOTICEs
    * (for OPERs only)
    *
    * Armin, 8Jun90 (gruner@informatik.tu-muenchen.de)
    */
-  if(*nick == '$')
+  if (*nick == '$')
   {
-    if((*(nick+1) == '$' || *(nick+1) == '#'))
+    if ((*(nick+1) == '$' || *(nick+1) == '#'))
       nick++;
     else if(MyOper(source_p))
     {
@@ -819,83 +944,35 @@ handle_opers(int p_or_n, char *command, struct Client *client_p,
 		 me.name, source_p->name, command, nick, nick);
       return;
     }
-      
+
     if ((s = strrchr(nick, '.')) == NULL)
     {
       sendto_one(source_p, form_str(ERR_NOTOPLEVEL),
                  me.name, source_p->name, nick);
       return;
     }
+
     while (*++s)
       if (*s == '.' || *s == '*' || *s == '?')
         break;
+
     if (*s == '*' || *s == '?')
     {
       sendto_one(source_p, form_str(ERR_WILDTOPLEVEL),
-                 me.name, source_p->name, nick);
+                 ID_or_name(&me, client_p),
+                 ID_or_name(source_p, client_p), nick);
       return;
     }
     
     sendto_match_butone(IsServer(client_p) ? client_p : NULL, source_p,
-                        nick + 1,
-                        (*nick == '#') ? MATCH_HOST : MATCH_SERVER,
+                        nick + 1, (*nick == '#') ? MATCH_HOST : MATCH_SERVER,
                         "%s $%s :%s", command, nick, text);
+
+    if ((p_or_n != NOTICE) && MyClient(source_p))
+      source_p->localClient->last = CurrentTime;
+
     return;
   }
-
-  /*
-   * user[%host]@server addressed?
-   */
-  if ((server = strchr(nick, '@')) && (target_p = find_server(server + 1)))
-  {
-    count = 0;
-
-    /*
-     * Not destined for a user on me :-(
-     */
-
-    if (!IsMe(target_p))
-    {
-      sendto_one(target_p, ":%s %s %s :%s", source_p->name,
-                 "PRIVMSG", nick, text);
-      return;
-    }
-
-    *server = '\0';
-
-    if ((host = strchr(nick, '%')) != NULL)
-      *host++ = '\0';
-
-    /* Check if someones msg'ing opers@our.server */
-    if (strcmp(nick, "opers") == 0)
-    {
-      sendto_realops_flags(FLAGS_ALL, L_ALL, "To opers: From: %s: %s",
-                           source_p->name, text);
-      return;
-    }
-
-    /*
-     * Look for users which match the destination host
-     * (no host == wildcard) and if one and one only is
-     * found connected to me, deliver message!
-     */
-    target_p = find_userhost(nick, host, &count);
-
-    if (target_p != NULL)
-    {
-      if (server != NULL)
-	*server = '@';
-      if (host != NULL)
-	*--host = '%';
-
-      if (count == 1)
-        sendto_anywhere(target_p, source_p, "%s %s :%s", "PRIVMSG",
-			nick, text);
-      else
-        sendto_one(source_p, form_str(ERR_TOOMANYTARGETS),
-                   me.name, source_p->name, nick);
-    }
-  }
 }
 
 /*
@@ -913,21 +990,27 @@ find_userhost(char *user, char *host, int *count)
 {
   struct Client *c2ptr;
   struct Client *res = NULL;
+  dlink_node *lc2ptr;
 
   *count = 0;
+
   if (collapse(user) != NULL)
+  {
+    DLINK_FOREACH(lc2ptr, local_client_list.head)
     {
-      for (c2ptr = GlobalClientList; c2ptr; c2ptr = c2ptr->next) 
-	{
-	  if (!MyClient(c2ptr)) /* implies mine and an user */
-	    continue;
-	  if ((!host || match(host, c2ptr->host)) &&
-	      irccmp(user, c2ptr->username) == 0)
-	    {
-	      (*count)++;
-	      res = c2ptr;
-	    }
-	}
+      c2ptr = lc2ptr->data;
+
+      if (!IsClient(c2ptr)) /* something other than a client */
+        continue;
+
+      if ((!host || match(host, c2ptr->host)) &&
+          irccmp(user, c2ptr->username) == 0)
+      {
+        (*count)++;
+        res = c2ptr;
+      }
     }
-  return (res);
+  }
+
+  return(res);
 }
