@@ -2,7 +2,7 @@
  *  ircd-hybrid: an advanced Internet Relay Chat Daemon(ircd).
  *  hostmask.c: Code to efficiently find IP & hostmask based configs.
  *
- *  Copyright (C) 2002 by the past and present ircd coders, and others.
+ *  Copyright (C) 2005 by the past and present ircd coders, and others.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,12 +19,13 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
  *  USA
  *
- *  $Id: hostmask.c,v 1.1.1.1 2002/09/12 01:09:20 cry0gen Exp $
+ *  $Id: hostmask.c 803 2007-02-16 09:31:25Z stu $
  */
 
 #include "stdinc.h"
 #include "memory.h"
 #include "ircd_defs.h"
+#include "tools.h"
 #include "s_conf.h"
 #include "hostmask.h"
 #include "numeric.h"
@@ -32,11 +33,11 @@
 #include "irc_string.h"
 
 #ifdef IPV6
-static int try_parse_v6_netmask(const char *, struct irc_inaddr *, int *);
-static unsigned long hash_ipv6(struct irc_inaddr *, int);
+static int try_parse_v6_netmask(const char *, struct irc_ssaddr *, int *);
+static unsigned long hash_ipv6(struct irc_ssaddr *, int);
 #endif
-static int try_parse_v4_netmask(const char *, struct irc_inaddr *, int *);
-static unsigned long hash_ipv4(struct irc_inaddr *, int);
+static int try_parse_v4_netmask(const char *, struct irc_ssaddr *, int *);
+static unsigned long hash_ipv4(struct irc_ssaddr *, int);
 
 #define DigitParse(ch) do { \
                        if (ch >= '0' && ch <= '9') \
@@ -49,7 +50,7 @@ static unsigned long hash_ipv4(struct irc_inaddr *, int);
 
 /* The mask parser/type determination code... */
 
-/* int try_parse_v6_netmask(const char *, struct irc_inaddr *, int *);
+/* int try_parse_v6_netmask(const char *, struct irc_ssaddr *, int *);
  * Input: A possible IPV6 address as a string.
  * Output: An integer describing whether it is an IPV6 or hostmask,
  *         an address(if it is IPV6), a bitlength(if it is IPV6).
@@ -62,7 +63,7 @@ static unsigned long hash_ipv4(struct irc_inaddr *, int);
 */
 #ifdef IPV6
 static int
-try_parse_v6_netmask(const char *text, struct irc_inaddr *addr, int *b)
+try_parse_v6_netmask(const char *text, struct irc_ssaddr *addr, int *b)
 {
   const char *p;
   char c;
@@ -73,6 +74,7 @@ try_parse_v6_netmask(const char *text, struct irc_inaddr *addr, int *b)
   int bits = 128;
   int deficit = 0;
   short dc[8];
+  struct sockaddr_in6 *v6 = (struct sockaddr_in6*) addr;
 
   for (p = text; (c = *p); p++)
     if (IsXDigit(c))
@@ -150,14 +152,14 @@ try_parse_v6_netmask(const char *text, struct irc_inaddr *addr, int *b)
   if (addr)
     for (dp = 0; dp < 8; dp++)
       /* The cast is a kludge to make netbsd work. */
-      ((unsigned short *)&addr->sins.sin6)[dp] = htons(dc[dp]);
+      ((unsigned short *)&v6->sin6_addr)[dp] = htons(dc[dp]);
   if (b != NULL)
     *b = bits;
   return HM_IPV6;
 }
 #endif
 
-/* int try_parse_v4_netmask(const char *, struct irc_inaddr *, int *);
+/* int try_parse_v4_netmask(const char *, struct irc_ssaddr *, int *);
  * Input: A possible IPV4 address as a string.
  * Output: An integer describing whether it is an IPV4 or hostmask,
  *         an address(if it is IPV4), a bitlength(if it is IPV4).
@@ -165,13 +167,14 @@ try_parse_v6_netmask(const char *text, struct irc_inaddr *addr, int *b)
  * Comments: Called from parse_netmask
  */
 static int
-try_parse_v4_netmask(const char *text, struct irc_inaddr *addr, int *b)
+try_parse_v4_netmask(const char *text, struct irc_ssaddr *addr, int *b)
 {
   const char *p;
   const char *digits[4];
   unsigned char addb[4];
   int n = 0, bits = 0;
   char c;
+  struct sockaddr_in *v4 = (struct sockaddr_in*) addr;
 
   digits[n++] = text;
 
@@ -194,8 +197,11 @@ try_parse_v4_netmask(const char *text, struct irc_inaddr *addr, int *b)
     else if (c == '/')
     {
       char *after;
+      errno = 0;
       bits = strtoul(p + 1, &after, 10);
-      if (!bits || *after)
+      if (*after)
+        return HM_HOST;
+      if(bits == 0 && errno != 0)
         return HM_HOST;
       if (bits > n * 8)
         return HM_HOST;
@@ -219,14 +225,14 @@ try_parse_v4_netmask(const char *text, struct irc_inaddr *addr, int *b)
   for (n = bits / 8 + (bits % 8 ? 1 : 0); n < 4; n++)
     addb[n] = 0;
   if (addr)
-    addr->sins.sin.s_addr =
+    v4->sin_addr.s_addr =
       htonl(addb[0] << 24 | addb[1] << 16 | addb[2] << 8 | addb[3]);
   if (b != NULL)
     *b = bits;
   return HM_IPV4;
 }
 
-/* int parse_netmask(const char *, struct irc_inaddr *, int *);
+/* int parse_netmask(const char *, struct irc_ssaddr *, int *);
  * Input: A hostmask, or an IPV4/6 address.
  * Output: An integer describing whether it is an IPV4, IPV6 address or a
  *         hostmask, an address(if it is an IP mask),
@@ -234,10 +240,10 @@ try_parse_v4_netmask(const char *text, struct irc_inaddr *addr, int *b)
  * Side effects: None
  */
 int
-parse_netmask(const char *text, struct irc_inaddr *addr, int *b)
+parse_netmask(const char *text, struct irc_ssaddr *addr, int *b)
 {
 #ifdef IPV6
-  if (strchr(text, ':'))
+    if (strchr(text, ':'))
     return try_parse_v6_netmask(text, addr, b);
 #endif
   if (strchr(text, '.'))
@@ -246,43 +252,88 @@ parse_netmask(const char *text, struct irc_inaddr *addr, int *b)
 }
 
 /* The address matching stuff... */
-/* int match_ipv6(struct irc_inaddr *, struct irc_inaddr *, int)
+/* int match_ipv6(struct irc_ssaddr *, struct irc_ssaddr *, int)
  * Input: An IP address, an IP mask, the number of bits in the mask.
  * Output: if match, -1 else 0
  * Side effects: None
  */
 #ifdef IPV6
 int
-match_ipv6(struct irc_inaddr *addr, struct irc_inaddr *mask, int bits)
+match_ipv6(struct irc_ssaddr *addr, struct irc_ssaddr *mask, int bits)
 {
   int i, m, n = bits / 8;
+  struct sockaddr_in6 *v6 = (struct sockaddr_in6*)addr;
+  struct sockaddr_in6 *v6mask = (struct sockaddr_in6*)mask;
 
   for (i = 0; i < n; i++)
-    if (addr->sins.sin6.s6_addr[i] != mask->sins.sin6.s6_addr[i])
+    if (v6->sin6_addr.s6_addr[i] != v6mask->sin6_addr.s6_addr[i])
       return 0;
   if ((m = bits % 8) == 0)
     return -1;
-  if ((addr->sins.sin6.s6_addr[n] & ~((1 << (8 - m)) - 1)) ==
-      mask->sins.sin6.s6_addr[n])
+  if ((v6->sin6_addr.s6_addr[n] & ~((1 << (8 - m)) - 1)) ==
+      v6mask->sin6_addr.s6_addr[n])
     return -1;
   return 0;
 }
 #endif
-
-/* int match_ipv4(struct irc_inaddr *, struct irc_inaddr *, int)
+/* int match_ipv4(struct irc_ssaddr *, struct irc_ssaddr *, int)
  * Input: An IP address, an IP mask, the number of bits in the mask.
  * Output: if match, -1 else 0
  * Side Effects: None
  */
 int
-match_ipv4(struct irc_inaddr *addr, struct irc_inaddr *mask, int bits)
+match_ipv4(struct irc_ssaddr *addr, struct irc_ssaddr *mask, int bits)
 {
-  if ((ntohl(addr->sins.sin.s_addr) & ~((1 << (32 - bits)) - 1)) !=
-      ntohl(mask->sins.sin.s_addr))
+  struct sockaddr_in *v4 = (struct sockaddr_in*) addr;
+  struct sockaddr_in *v4mask = (struct sockaddr_in*) mask;
+  if ((ntohl(v4->sin_addr.s_addr) & ~((1 << (32 - bits)) - 1)) !=
+      ntohl(v4mask->sin_addr.s_addr))
     return 0;
   return -1;
 }
 
+/*
+ * mask_addr
+ *
+ * inputs       - pointer to the ip to mask
+ *              - bitlen
+ * output       - NONE
+ * side effects -
+ */
+void
+mask_addr(struct irc_ssaddr *ip, int bits)
+{
+  int mask;
+#ifdef IPV6
+  struct sockaddr_in6 *v6_base_ip;
+  int i, m, n;
+#endif
+  struct sockaddr_in *v4_base_ip;
+
+#ifdef IPV6
+  if (ip->ss.ss_family != AF_INET6)
+#endif
+  {
+    v4_base_ip = (struct sockaddr_in*)ip;
+    mask = ~((1 << (32 - bits)) - 1);
+    v4_base_ip->sin_addr.s_addr =
+      htonl(ntohl(v4_base_ip->sin_addr.s_addr) & mask);
+  }
+#ifdef IPV6
+  else
+  {
+    n = bits / 8;
+    m = bits % 8;
+    v6_base_ip = (struct sockaddr_in6*)ip;
+
+    mask = ~((1 << (8 - m)) -1 );
+    v6_base_ip->sin6_addr.s6_addr[n] = v6_base_ip->sin6_addr.s6_addr[n] & mask;
+    for (i = n + 1; i < 16; i++)
+      v6_base_ip->sin6_addr.s6_addr[i] = 0;
+  }
+#endif
+}
+
 /* Hashtable stuff...now external as its used in m_stats.c */
 struct AddressRec *atable[ATABLE_SIZE];
 
@@ -292,38 +343,46 @@ init_host_hash(void)
   memset(&atable, 0, sizeof(atable));
 }
 
-/* unsigned long hash_ipv4(struct irc_inaddr*)
+/* unsigned long hash_ipv4(struct irc_ssaddr*)
  * Input: An IP address.
  * Output: A hash value of the IP address.
  * Side effects: None
  */
 static unsigned long
-hash_ipv4(struct irc_inaddr *addr, int bits)
+hash_ipv4(struct irc_ssaddr *addr, int bits)
 {
-  unsigned long av = ntohl(addr->sins.sin.s_addr) & ~((1 << (32 - bits)) - 1);
-  return (av ^ (av >> 12) ^ (av >> 24)) & (ATABLE_SIZE - 1);
+  if (bits != 0)
+  {
+    struct sockaddr_in *v4 = (struct sockaddr_in *)addr;
+    unsigned long av = ntohl(v4->sin_addr.s_addr) & ~((1 << (32 - bits)) - 1);
+    return((av ^ (av >> 12) ^ (av >> 24)) & (ATABLE_SIZE - 1));
+  }
+
+  return(0);
 }
 
-/* unsigned long hash_ipv6(struct irc_inaddr*)
+/* unsigned long hash_ipv6(struct irc_ssaddr*)
  * Input: An IP address.
  * Output: A hash value of the IP address.
  * Side effects: None
  */
 #ifdef IPV6
 static unsigned long
-hash_ipv6(struct irc_inaddr *addr, int bits)
+hash_ipv6(struct irc_ssaddr *addr, int bits)
 {
   unsigned long v = 0, n;
+  struct sockaddr_in6 *v6 = (struct sockaddr_in6*) addr;
+  
   for (n = 0; n < 16; n++)
   {
     if (bits >= 8)
     {
-      v ^= addr->sins.sin6.s6_addr[n];
+      v ^= v6->sin6_addr.s6_addr[n];
       bits -= 8;
     }
     else if (bits)
     {
-      v ^= addr->sins.sin6.s6_addr[n] & ~((1 << (8 - bits)) - 1);
+      v ^= v6->sin6_addr.s6_addr[n] & ~((1 << (8 - bits)) - 1);
       return v & (ATABLE_SIZE - 1);
     }
     else
@@ -363,32 +422,36 @@ get_mask_hash(const char *text)
   const char *hp = "", *p;
 
   for (p = text + strlen(text) - 1; p >= text; p--)
-    if (*p == '*' || *p == '?')
+    if (IsMWildChar(*p))
       return hash_text(hp);
     else if (*p == '.')
       hp = p + 1;
   return hash_text(text);
 }
 
-/* struct ConfItem* find_conf_by_address(const char*, struct irc_inaddr*,
- *         int type, int fam, const char *username)
+/* struct AccessItem *find_conf_by_address(const char *, struct irc_ssaddr *,
+ *                                         int type, int fam, const char *username)
  * Input: The hostname, the address, the type of mask to find, the address
  *        family, the username.
  * Output: The matching value with the highest precedence.
  * Side-effects: None
  * Note: Setting bit 0 of the type means that the username is ignored.
+ * Warning: IsNeedPassword for everything that is not an auth{} entry
+ * should always be true (i.e. aconf->flags & CONF_FLAGS_NEED_PASSWORD == 0)
  */
-struct ConfItem *
-find_conf_by_address(const char *name, struct irc_inaddr *addr, int type,
-                     int fam, const char *username)
+struct AccessItem *
+find_conf_by_address(const char *name, struct irc_ssaddr *addr, int type,
+                     int fam, const char *username, const char *password)
 {
   unsigned long hprecv = 0;
-  struct ConfItem *hprec = NULL;
+  struct AccessItem *hprec = NULL;
   struct AddressRec *arec;
   int b;
 
   if (username == NULL)
     username = "";
+  if (password == NULL)
+    password = "";
 
   if (addr)
   {
@@ -400,11 +463,13 @@ find_conf_by_address(const char *name, struct irc_inaddr *addr, int type,
       {
         for (arec = atable[hash_ipv6(addr, b)]; arec; arec = arec->next)
           if (arec->type == (type & ~0x1) &&
+              arec->precedence > hprecv &&
               arec->masktype == HM_IPV6 &&
               match_ipv6(addr, &arec->Mask.ipa.addr,
                          arec->Mask.ipa.bits) &&
               (type & 0x1 || match(arec->username, username)) &&
-              arec->precedence > hprecv)
+	      (IsNeedPassword(arec->aconf) || arec->aconf->passwd == NULL ||
+	       match_conf_password(password, arec->aconf)))
           {
             hprecv = arec->precedence;
             hprec = arec->aconf;
@@ -419,11 +484,13 @@ find_conf_by_address(const char *name, struct irc_inaddr *addr, int type,
       {
         for (arec = atable[hash_ipv4(addr, b)]; arec; arec = arec->next)
           if (arec->type == (type & ~0x1) &&
+              arec->precedence > hprecv &&
               arec->masktype == HM_IPV4 &&
               match_ipv4(addr, &arec->Mask.ipa.addr,
                          arec->Mask.ipa.bits) &&
               (type & 0x1 || match(arec->username, username)) &&
-              arec->precedence > hprecv)
+	      (IsNeedPassword(arec->aconf) || arec->aconf->passwd == NULL ||
+	       match_conf_password(password, arec->aconf)))
           {
             hprecv = arec->precedence;
             hprec = arec->aconf;
@@ -434,104 +501,158 @@ find_conf_by_address(const char *name, struct irc_inaddr *addr, int type,
 
   if (name != NULL)
   {
-    const char *p;
-    /* And yes - we have to check p after strchr and p after increment for
-     * NULL -kre */
-    for (p = name; p != NULL; )
+    const char *p = name;
+
+    while (1)
     {
-      for (arec = atable[hash_text(p)]; arec; arec = arec->next)
+      for (arec = atable[hash_text(p)]; arec != NULL; arec = arec->next)
         if ((arec->type == (type & ~0x1)) &&
+            arec->precedence > hprecv &&
             (arec->masktype == HM_HOST) &&
             match(arec->Mask.hostname, name) &&
             (type & 0x1 || match(arec->username, username)) &&
-            (arec->precedence > hprecv))
+            (IsNeedPassword(arec->aconf) || arec->aconf->passwd == NULL ||
+             match_conf_password(password, arec->aconf)))
         {
           hprecv = arec->precedence;
           hprec = arec->aconf;
         }
-        p = strchr(p, '.');
-        if (p != NULL)
-          p++;
-        else
-          break;
+      p = strchr(p, '.');
+      if (p == NULL)
+        break;
+      p++;
     }
     for (arec = atable[0]; arec; arec = arec->next)
       if (arec->type == (type & ~0x1) &&
+          arec->precedence > hprecv &&
           arec->masktype == HM_HOST &&
           match(arec->Mask.hostname, name) &&
           (type & 0x1 || match(arec->username, username)) &&
-          arec->precedence > hprecv)
+          (IsNeedPassword(arec->aconf) || arec->aconf->passwd == NULL ||
+           match_conf_password(password, arec->aconf)))
       {
         hprecv = arec->precedence;
         hprec = arec->aconf;
       }
   }
+
   return hprec;
 }
 
-/* struct ConfItem* find_address_conf(const char*, const char*,
- * 	                               struct irc_inaddr*, int);
+/* struct AccessItem* find_address_conf(const char*, const char*,
+ * 	                               struct irc_ssaddr*, int, char *);
  * Input: The hostname, username, address, address family.
- * Output: The applicable ConfItem.
+ * Output: The applicable AccessItem.
  * Side-effects: None
  */
-struct ConfItem *
+struct AccessItem *
 find_address_conf(const char *host, const char *user,
-                  struct irc_inaddr *ip, int aftype)
+                  struct irc_ssaddr *ip, int aftype, char *password)
 {
-  struct ConfItem *iconf, *kconf;
+  struct AccessItem *iconf, *kconf;
 
   /* Find the best I-line... If none, return NULL -A1kmm */
-  if (!(iconf = find_conf_by_address(host, ip, CONF_CLIENT, aftype, user)))
-    return NULL;
+  if ((iconf = find_conf_by_address(host, ip, CONF_CLIENT, aftype, user,
+                                    password)) == NULL)
+    return(NULL);
 
   /* If they are exempt from K-lines, return the best I-line. -A1kmm */
   if (IsConfExemptKline(iconf))
-    return iconf;
+    return(iconf);
 
   /* Find the best K-line... -A1kmm */
-  kconf = find_conf_by_address(host, ip, CONF_KILL, aftype, user);
+  kconf = find_conf_by_address(host, ip, CONF_KILL, aftype, user, NULL);
 
   /* If they are K-lined, return the K-line. Otherwise, return the
    * I-line. -A1kmm */
-  if (kconf)
-    return kconf;
-  return iconf;
+  if (kconf != NULL)
+    return(kconf);
+
+  kconf = find_conf_by_address(host, ip, CONF_GLINE, aftype, user, NULL);
+  if (kconf != NULL && !IsConfExemptGline(iconf))
+    return(kconf);
+
+  return(iconf);
 }
 
-/* struct ConfItem* find_dline(struct irc_inaddr*, int)
- * Input: An address, an address family.
- * Output: The best matching D-line or exempt line.
+struct AccessItem *
+find_gline_conf(const char *host, const char *user,
+                struct irc_ssaddr *ip, int aftype)
+{
+  struct AccessItem *eline;
+
+  eline = find_conf_by_address(host, ip, CONF_EXEMPTKLINE, aftype,
+                               user, NULL);
+  if (eline != NULL)
+    return(eline);
+
+  return(find_conf_by_address(host, ip, CONF_GLINE, aftype, user, NULL));
+}
+
+/* find_kline_conf
+ *
+ * inputs	- pointer to hostname
+ *		- pointer to username
+ *		- incoming IP and type (IPv4 vs. IPv6)
+ * outut	- pointer to kline conf if found NULL if not
+ * side effects	-
+ */
+struct AccessItem *
+find_kline_conf(const char *host, const char *user,
+		struct irc_ssaddr *ip, int aftype)
+{
+  struct AccessItem *eline;
+
+  eline = find_conf_by_address(host, ip, CONF_EXEMPTKLINE, aftype,
+                               user, NULL);
+  if (eline != NULL)
+    return(eline);
+
+  return(find_conf_by_address(host, ip, CONF_KILL, aftype, user, NULL));
+}
+
+/* struct AccessItem* find_dline_conf(struct irc_ssaddr*, int)
+ *
+ * Input:	An address, an address family.
+ * Output:	The best matching D-line or exempt line.
  * Side effects: None.
  */
-struct ConfItem *
-find_dline(struct irc_inaddr *addr, int aftype)
+struct AccessItem *
+find_dline_conf(struct irc_ssaddr *addr, int aftype)
 {
-  struct ConfItem *eline;
+  struct AccessItem *eline;
+
   eline = find_conf_by_address(NULL, addr, CONF_EXEMPTDLINE | 1, aftype,
-                               NULL);
-  if (eline)
-    return eline;
-  return find_conf_by_address(NULL, addr, CONF_DLINE | 1, aftype, NULL);
+                               NULL, NULL);
+  if (eline != NULL)
+    return(eline);
+  return(find_conf_by_address(NULL, addr, CONF_DLINE | 1, aftype, NULL, NULL));
 }
 
-/* void add_conf_by_address(const char*, int, const char *,
- *         struct ConfItem *aconf)
+/* void add_conf_by_address(int, struct AccessItem *aconf)
  * Input: 
  * Output: None
  * Side-effects: Adds this entry to the hash table.
  */
 void
-add_conf_by_address(const char *address, int type, const char *username,
-                    struct ConfItem *aconf)
+add_conf_by_address(int type, struct AccessItem *aconf)
 {
+  const char *address;
+  const char *username;
   static unsigned long prec_value = 0xFFFFFFFF;
   int masktype, bits;
   unsigned long hv;
   struct AddressRec *arec;
 
-  if (address == NULL)
+  address = aconf->host;
+  username = aconf->user;
+
+  assert(type != 0); 
+  assert(aconf != NULL);
+
+  if (EmptyString(address))
     address = "/NOMATCH!/";
+
   arec = MyMalloc(sizeof(struct AddressRec));
   masktype = parse_netmask(address, &arec->Mask.ipa.addr, &bits);
   arec->Mask.ipa.bits = bits;
@@ -544,7 +665,7 @@ add_conf_by_address(const char *address, int type, const char *username,
     arec->next = atable[(hv = hash_ipv6(&arec->Mask.ipa.addr, bits))];
     atable[hv] = arec;
   }
-  else
+  else 
 #endif
   if (masktype == HM_IPV4)
   {
@@ -565,20 +686,21 @@ add_conf_by_address(const char *address, int type, const char *username,
   arec->type = type;
 }
 
-/* void delete_one_address(const char*, struct ConfItem*)
- * Input: An address string, the associated ConfItem.
+/* void delete_one_address(const char*, struct AccessItem*)
+ * Input: An address string, the associated AccessItem.
  * Output: None
- * Side effects: Deletes an address record. Frees the ConfItem if there
+ * Side effects: Deletes an address record. Frees the AccessItem if there
  *               is nothing referencing it, sets it as illegal otherwise.
  */
 void
-delete_one_address_conf(const char *address, struct ConfItem *aconf)
+delete_one_address_conf(const char *address, struct AccessItem *aconf)
 {
   int masktype, bits;
   unsigned long hv;
   struct AddressRec *arec, *arecl = NULL;
-  struct irc_inaddr addr;
+  struct irc_ssaddr addr;
   masktype = parse_netmask(address, &addr, &bits);
+
 #ifdef IPV6
   if (masktype == HM_IPV6)
   {
@@ -604,9 +726,9 @@ delete_one_address_conf(const char *address, struct ConfItem *aconf)
         arecl->next = arec->next;
       else
         atable[hv] = arec->next;
-      aconf->flags |= CONF_ILLEGAL;
-      if (!aconf->clients)
-        free_conf(aconf);
+      aconf->status |= CONF_ILLEGAL;
+      if (aconf->clients == 0)
+        free_access_item(aconf);
       MyFree(arec);
       return;
     }
@@ -618,55 +740,66 @@ delete_one_address_conf(const char *address, struct ConfItem *aconf)
  * Input: None
  * Output: None
  * Side effects: Clears out all address records in the hash table,
- *               frees them, and frees the ConfItems if nothing references
- *               them, otherwise sets them as illegal.
+ *               frees them, and frees the AccessItems if nothing references
+ *               them, otherwise sets them as illegal.   
  */
 void
 clear_out_address_conf(void)
 {
   int i;
-  struct AddressRec **store_next;
-  struct AddressRec *arec, *arecn;
-
+  struct AddressRec *arec;
+  struct AddressRec *last_arec;
+  struct AddressRec *next_arec;
+ 
   for (i = 0; i < ATABLE_SIZE; i++)
   {
-    store_next = &atable[i];
-    for (arec = atable[i]; arec; arec = arecn)
+    last_arec = NULL;
+    for (arec = atable[i]; arec; arec = next_arec)
     {
-      arecn = arec->next;
       /* We keep the temporary K-lines and destroy the
-       * permanent ones, just to be confusing :) -A1kmm */
+       * permanent ones, just to be confusing :) -A1kmm
+       */  
+      next_arec = arec->next;
+
       if (arec->aconf->flags & CONF_FLAGS_TEMPORARY)
       {
-        *store_next = arec;
-        store_next = &arec->next;
+        last_arec = arec;
       }
-      else
+      else   
       {
-        arec->aconf->flags |= CONF_ILLEGAL;
-        if (!arec->aconf->clients)
-          free_conf(arec->aconf);
+        /* unlink it from link list - Dianora */
+  
+        if (last_arec == NULL)
+        {
+          atable[i] = next_arec;
+        }
+        else
+        {
+          last_arec->next = next_arec;
+        }
+
+        arec->aconf->status |= CONF_ILLEGAL;
+        if (arec->aconf->clients == 0)
+          free_access_item(arec->aconf);
         MyFree(arec);
       }
     }
-    *store_next = NULL;
   }
 }
 
-
 /*
  * show_iline_prefix()
  *
  * inputs       - pointer to struct Client requesting output
- *              - pointer to struct ConfItem 
+ *              - pointer to struct AccessItem 
  *              - name to which iline prefix will be prefixed to
  * output       - pointer to static string with prefixes listed in ascii form
  * side effects - NONE
  */
 char *
-show_iline_prefix(struct Client *sptr, struct ConfItem *aconf, char *name)
+show_iline_prefix(struct Client *sptr, struct AccessItem *aconf, const char *name)
 {
-  static char prefix_of_host[USERLEN + 15];
+  static char prefix_of_host[USERLEN + 14];
   char *prefix_ptr;
 
   prefix_ptr = prefix_of_host;
@@ -676,7 +809,9 @@ show_iline_prefix(struct Client *sptr, struct ConfItem *aconf, char *name)
     *prefix_ptr++ = '!';
   if (IsNeedIdentd(aconf))
     *prefix_ptr++ = '+';
-  if (IsPassIdentd(aconf))
+  if (!IsNeedPassword(aconf))
+    *prefix_ptr++ = '&';
+  if (IsConfExemptResv(aconf))
     *prefix_ptr++ = '$';
   if (IsNoMatchIp(aconf))
     *prefix_ptr++ = '%';
@@ -684,13 +819,17 @@ show_iline_prefix(struct Client *sptr, struct ConfItem *aconf, char *name)
     *prefix_ptr++ = '=';
   if (MyOper(sptr) && IsConfExemptKline(aconf))
     *prefix_ptr++ = '^';
+  if (MyOper(sptr) && IsConfExemptGline(aconf))
+    *prefix_ptr++ = '_';
   if (MyOper(sptr) && IsConfExemptLimits(aconf))
     *prefix_ptr++ = '>';
   if (MyOper(sptr) && IsConfIdlelined(aconf))
     *prefix_ptr++ = '<';
-  *prefix_ptr = '\0';
-  strncpy(prefix_ptr, name, USERLEN);
-  return (prefix_of_host);
+  if (IsConfCanFlood(aconf))
+    *prefix_ptr++ = '|';
+  strlcpy(prefix_ptr, name, USERLEN);
+
+  return(prefix_of_host);
 }
 
 /* report_auth()
@@ -702,13 +841,15 @@ show_iline_prefix(struct Client *sptr, struct ConfItem *aconf, char *name)
 void
 report_auth(struct Client *client_p)
 {
-  char *name, *host, *pass, *user, *classname;
   struct AddressRec *arec;
-  struct ConfItem *aconf;
-  int i, port;
+  struct ConfItem *conf;
+  struct AccessItem *aconf;
+  int i;
 
   for (i = 0; i < ATABLE_SIZE; i++)
+  {
     for (arec = atable[i]; arec; arec = arec->next)
+    {
       if (arec->type == CONF_CLIENT)
       {
         aconf = arec->aconf;
@@ -716,20 +857,30 @@ report_auth(struct Client *client_p)
         if (!MyOper(client_p) && IsConfDoSpoofIp(aconf))
           continue;
 
-        get_printable_conf(aconf, &name, &host, &pass, &user, &port,
-                           &classname);
+	conf = unmap_conf_item(aconf);
 
         /* We are doing a partial list, based on what matches the u@h of the
-         * sender, so prepare the strings for comparing --fl_ */
-
-        sendto_one(client_p, form_str(RPL_STATSILINE), me.name,
-                   client_p->name, (IsConfRestricted(aconf)) ? 'i' : 'I', name,
-                   show_iline_prefix(client_p, aconf, user),
-#ifdef HIDE_SPOOF_IPS
-                   IsConfDoSpoofIp(aconf) ? "255.255.255.255" :
-#endif
-                   host, port, classname);
+         * sender, so prepare the strings for comparing --fl_
+	 */
+        if (ConfigFileEntry.hide_spoof_ips)
+          sendto_one(client_p, form_str(RPL_STATSILINE), me.name,
+                     client_p->name, 'I',
+		     conf->name == NULL ? "*" : conf->name,
+		     show_iline_prefix(client_p, aconf, aconf->user),
+                     IsConfDoSpoofIp(aconf) ? "255.255.255.255" :
+                     aconf->host, aconf->port,
+		     aconf->class_ptr ? aconf->class_ptr->name : "<default>");
+		     
+        else
+          sendto_one(client_p, form_str(RPL_STATSILINE), me.name,
+                     client_p->name, 'I',
+		     conf->name == NULL ? "*" : conf->name,
+		     show_iline_prefix(client_p, aconf, aconf->user),
+                     aconf->host, aconf->port,
+		     aconf->class_ptr ? aconf->class_ptr->name : "<default>");
       }
+    }
+  }
 }
 
 /* report_Klines()
@@ -742,27 +893,35 @@ report_auth(struct Client *client_p)
 void
 report_Klines(struct Client *client_p, int tkline)
 {
-  char *name, *host, *pass, *user, *classname, c;
-  struct AddressRec *arec;
-  struct ConfItem *aconf = NULL;
-  int i, port;
+  struct AddressRec *arec = NULL;
+  struct AccessItem *aconf = NULL;
+  int i;
+  const char *p = NULL;
 
   if (tkline)
-    c = 'k';
+    p = "k";
   else
-    c = 'K';
+    p = "K";
 
   for (i = 0; i < ATABLE_SIZE; i++)
+  {
     for (arec = atable[i]; arec; arec = arec->next)
+    {
       if (arec->type == CONF_KILL)
       {
-        if ((tkline && !((aconf = arec->aconf)->flags & CONF_FLAGS_TEMPORARY))
-            || (!tkline
-                && ((aconf = arec->aconf)->flags & CONF_FLAGS_TEMPORARY)))
+        if ((tkline && !((aconf = arec->aconf)->flags & CONF_FLAGS_TEMPORARY)) ||
+            (!tkline && ((aconf = arec->aconf)->flags & CONF_FLAGS_TEMPORARY)))
           continue;
-        get_printable_conf(aconf, &name, &host, &pass, &user, &port,
-                           &classname);
-        sendto_one(client_p, form_str(RPL_STATSKLINE), me.name,
-                   client_p->name, c, host, user, pass);
+
+	if (IsOper(client_p))
+	  sendto_one(client_p, form_str(RPL_STATSKLINE), me.name,
+                     client_p->name, p, aconf->host, aconf->user,
+		     aconf->reason, aconf->oper_reason ? aconf->oper_reason : "");
+	else
+          sendto_one(client_p, form_str(RPL_STATSKLINE), me.name,
+                     client_p->name, p, aconf->host, aconf->user,
+		     aconf->reason, "");
       }
+    }
+  }
 }
